-- | Calculator User interface with Threepenny GUI
module ThreepennyGUI(runCalculator,Calculator(..)) where
import System.IO.Error(tryIOError)

import qualified Graphics.UI.Threepenny as UI
import Graphics.UI.Threepenny.Core as UI hiding (column,row)

import ThreepennyPages
import ThreepennyPictures

runCalculator :: Calculator expr -> IO ()
runCalculator = startGUI defaultConfig . setup

-- | To run the user interface, the calculator implementation should
-- provide the following functions
data Calculator expr =
     Calc { readExpr :: String -> Maybe expr,
            showExpr :: expr -> String,
            diffExpr :: expr -> expr,
            drawGraph :: (Double,Double) -> Double -> expr -> Picture,
                      -- canvas size,       scale
                      -- (width,height)     5..200
            loadExpr :: FilePath -> IO expr,
            saveExpr :: FilePath -> expr -> IO ()
          }

canWidth,canHeight :: Num a => a
canWidth  = 600
canHeight = 400
canSize = (canWidth,canHeight)

setup :: Calculator expr -> Window -> UI ()
setup calc window =
 do pure window # set title "Calculator with ThreepennyGUI"
    -- Elements
    hdr     <- UI.h1 #+ [string "Calculator"]
    canvas  <- mkCanvas canWidth canHeight   -- The drawing area
    fx      <- mkHTML "<i>f</i>(<i>x</i>)="  -- The text "f(x)="
    input   <- mkInput 30 "x"                -- The formula input
    drawBtn <- mkButton "Draw graph"         -- The draw button
    diff    <- mkButton "Differentiate"      -- The differentiate button
    msg     <- UI.span # set style [("color","red")] -- Error messages
      -- The markup "<i>...</i>" means that the text inside should be rendered
      -- in italics.

    filename <- mkInput 20 "function.txt"
    load <- mkButton "Load"
    save <- mkButton "Save"
    filemsg <- UI.span # set style [("color","red")] -- File error messages

    let z0 :: Num a => a
        z0 = 50
    zoom <- mkSlider (5,200) z0

    -- Layout
    formula <- UI.div #+ [pure fx,pure input,pure msg]
    buttons <- UI.div #+ [pure drawBtn,pure diff]

    getBody window #+
        [pure hdr,
         pure canvas,
         UI.div #+ [pure zoom],
         pure formula,
         pure buttons,
         UI.paragraph #+ [mkHTML "Filename: ",
                          pure filename,pure load,pure save,
                          pure filemsg]]

    -- Styling
    getBody window # set style [("backgroundColor","lightblue"),
                                ("textAlign","center")]
    pure input # set style [("fontSize","inherit")]
--  focus input
--  select input

    -- Interaction

    let s0 = (readExpr calc "x",z0)
    let inputE0 = valueChange' input
    (inputE,trigger) <- liftIO newEvent
    let setInput s = set' value s input >> liftIO (trigger s)
    onEvent inputE0 (liftIO . trigger)
    inputB <- fmap (readExpr calc) <$> stepper "x" inputE
    zoomB <- stepper z0 (read <$> valueChange' zoom)

    on UI.click drawBtn $ \ _ -> do s <- get value input
                                    liftIO (trigger s)
    on UI.click diff $ \ _ ->
                do i <- currentValue inputB
                   case i of
                     Nothing -> return ()
                     Just e ->
                       do let e' = diffExpr calc e
                              s = showExpr calc e'
                          setInput s

    let graphB = (,) <$> inputB <*> zoomB

        redrawGraph (Just e,k) =
              do pure msg # set UI.html ""
                 draw (drawGraph calc canSize k e) canSize canvas
        redrawGraph _ =
              do pure msg # set UI.html "Syntax error!"
                 draw blank canSize canvas

    on UI.click save $ \ _ ->
        do name <- get value filename
           i <- currentValue inputB
           case i of
             Just e -> do r <- liftIO (tryIOError (saveExpr calc name e))
                          let msg = either show (const "") r
                          filemsg # set' UI.text msg
             _ -> filemsg # set' UI.text "syntax error, can't save"

    on UI.click load $ \ _ ->
        do name <- get value filename
           e <- liftIO (tryIOError (loadExpr calc name))
           case e of
             Left e -> filemsg # set' UI.text (show e)
             Right e -> do filemsg # set' UI.text ""
                           setInput (showExpr calc e)

    onChanges graphB redrawGraph

    redrawGraph s0 -- draw the initial function
